# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Wrapper of BFGS algorithm.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from typing import Sequence, Set
cimport cython
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport M_PI, cos, sin
from libcpp.pair cimport pair
from .sketch_solve cimport (
    point_on_point,
    p2p_distance,
    internal_angle,
    point_on_line,
    line_internal_angle,
    solve,
)
from .expression cimport get_vlinks, VJoint, VPoint, Coord


cdef inline double *de_refer_post_inc(clist[double].iterator &it):
    """Implement &(*it++) in C++."""
    return &cython.operator.dereference(cython.operator.postincrement(it))


cdef inline void _sort_pairs(dict data_dict):
    """Sort the pairs in data_dict."""
    for k in data_dict:
        if isinstance(k, (Sequence, Set)) and len(k) == 2:
            data_dict[frozenset(k)] = data_dict.pop(k)


cdef inline double _radians(double degree):
    """Degrees to radians."""
    return degree / 180 * M_PI


cdef class SolverSystem:
    """Sketch Solve solver.

    !!! note
        The object attributes of such type are unable to access.
    """

    def __cinit__(self, object vpoints_, dict inputs = None, dict data_dict = None):
        self.vpoints = list(vpoints_)
        self.vlinks = {vlink.name: vlink for vlink in get_vlinks(self.vpoints)}
        self.inputs = {} if inputs is None else inputs
        self.data_dict = {} if data_dict is None else data_dict
        _sort_pairs(self.data_dict)
        self.build_expression()

    cpdef bint same_points(self, object vpoints_):
        """Return true if two expressions are same."""
        cdef int i
        cdef VPoint p1, p2
        for i, p1 in enumerate(vpoints_):
            p2 = self.vpoints[i]
            if p1.links != p2.links:
                return False
        return True

    cpdef frozenset show_inputs(self):
        """Show the current input pairs keys from original constructor."""
        return frozenset(self.inputs)

    cpdef frozenset show_data(self):
        """Show the current keys of `data_dict` parameter from original
        constructor.
        """
        return frozenset(self.data_dict)

    cdef Point *point_ptr(self, int i, VLink vlink):
        """Pick Point pointers."""
        cdef VPoint vp = self.vpoints[i]
        if i not in self.data_dict and vp.is_slot_link(vlink.name):
            return &self.slider_bases[self.sliders[i]]
        else:
            return &self.points[i]

    cdef void build_expression(self):
        """Build the expression for solver at first time."""
        # Point parameters
        cdef int i
        cdef double x, y
        cdef double *tmp_ptr
        cdef VPoint vpoint
        cdef Coord coord
        for i, vpoint in enumerate(self.vpoints):
            if vpoint.no_link():
                x = vpoint.c[0, 0]
                y = vpoint.c[0, 1]
                self.constants.push_back(x)
                tmp_ptr = &self.constants.back()
                self.constants.push_back(y)
                self.points.push_back([tmp_ptr, &self.constants.back()])
                continue
            if vpoint.grounded():
                if self.check_known(i):
                    continue
                x = vpoint.c[0, 0]
                y = vpoint.c[0, 1]
                self.constants.push_back(x)
                tmp_ptr = &self.constants.back()
                self.constants.push_back(y)
                if vpoint.type in {VJoint.P, VJoint.RP}:
                    self.sliders[i] = <int>self.slider_bases.size()
                    # Base point (slot) is fixed
                    self.slider_bases.push_back([tmp_ptr, &self.constants.back()])
                    # Slot point (slot) is movable
                    self.params.push_back(x + cos(vpoint.angle))
                    tmp_ptr = &self.params.back()
                    self.params.push_back(y + sin(vpoint.angle))
                    self.slider_slots.push_back([tmp_ptr, &self.params.back()])
                    # Pin is movable
                    x = vpoint.c[1, 0]
                    y = vpoint.c[1, 1]
                    if vpoint.has_offset() and vpoint.true_offset() <= 0.1:
                        if vpoint.offset() > 0:
                            x += 0.1
                            y += 0.1
                        else:
                            x -= 0.1
                            y -= 0.1
                    self.params.push_back(x)
                    tmp_ptr = &self.params.back()
                    self.params.push_back(y)
                    self.points.push_back([tmp_ptr, &self.params.back()])
                else:
                    self.points.push_back([tmp_ptr, &self.constants.back()])
                continue
            if self.check_known(i):
                continue
            x = vpoint.c[0, 0]
            y = vpoint.c[0, 1]
            self.params.push_back(x)
            tmp_ptr = &self.params.back()
            self.params.push_back(y)
            if vpoint.type in {VJoint.P, VJoint.RP}:
                self.sliders[i] = <int>self.slider_bases.size()
                # Base point (slot) is movable
                self.slider_bases.push_back([tmp_ptr, &self.params.back()])
                # Slot point (slot) is movable
                self.params.push_back(x + cos(vpoint.angle))
                tmp_ptr = &self.params.back()
                self.params.push_back(y + sin(vpoint.angle))
                self.slider_slots.push_back([tmp_ptr, &self.params.back()])
                if vpoint.pin_grounded():
                    # Pin is fixed
                    x = vpoint.c[1, 0]
                    y = vpoint.c[1, 1]
                    self.constants.push_back(x)
                    tmp_ptr = &self.constants.back()
                    self.constants.push_back(y)
                    self.points.push_back([tmp_ptr, &self.constants.back()])
                else:
                    # Pin is movable
                    x = vpoint.c[1, 0]
                    y = vpoint.c[1, 1]
                    if vpoint.has_offset() and vpoint.true_offset() <= 0.1:
                        if vpoint.offset() > 0:
                            x += 0.1
                            y += 0.1
                        else:
                            x -= 0.1
                            y -= 0.1
                    self.params.push_back(x)
                    tmp_ptr = &self.params.back()
                    self.params.push_back(y)
                    self.points.push_back([tmp_ptr, &self.params.back()])
                continue
            # Point is movable
            self.points.push_back([tmp_ptr, &self.params.back()])
        # Link constraints
        # (automatic fill up the link length options of data keys)
        cdef int a, b, c, d
        cdef VPoint vp1, vp2
        cdef Point *p1
        cdef Point *p2
        cdef VLink vlink
        for vlink in self.vlinks.values():
            if len(vlink.points) < 2 or vlink.name == VLink.FRAME:
                continue
            a = vlink.points[0]
            b = vlink.points[1]
            if (a not in self.data_dict) or (b not in self.data_dict):
                vp1 = self.vpoints[a]
                vp2 = self.vpoints[b]
                p1 = self.point_ptr(a, vlink)
                p2 = self.point_ptr(b, vlink)
                frozen_pair = frozenset({a, b})
                if frozen_pair in self.data_dict:
                    x = self.data_dict[frozen_pair]
                else:
                    x = vp1.distance(vp2)
                    self.data_dict[frozen_pair] = x
                self.data_values.push_back(x)
                self.cons_list.push_back(p2p_distance(p1, p2, &self.data_values.back()))
            for c in vlink.points[2:]:
                if c in self.data_dict:
                    # Known coordinate
                    continue
                for d in (a, b):
                    vp1 = self.vpoints[c]
                    vp2 = self.vpoints[d]
                    p1 = self.point_ptr(c, vlink)
                    p2 = self.point_ptr(d, vlink)
                    frozen_pair = frozenset({c, d})
                    if frozen_pair in self.data_dict:
                        x = self.data_dict[frozen_pair]
                    else:
                        x = vp1.distance(vp2)
                        self.data_dict[frozen_pair] = x
                    self.data_values.push_back(x)
                    self.cons_list.push_back(p2p_distance(p1, p2, &self.data_values.back()))
        # Slider constraints
        cdef Line *slider_slot
        cdef pair[int, int] slider
        for slider in self.sliders:
            a = slider.first
            b = slider.second
            # Base point
            vp1 = self.vpoints[a]
            p1 = &self.points[a]
            # Base slot
            self.slider_lines.push_back([&self.slider_bases[b], &self.slider_slots[b]])
            slider_slot = &self.slider_lines.back()
            if vp1.grounded():
                # Slot is grounded
                self.constants.push_back(_radians(vp1.angle))
                self.cons_list.push_back(line_internal_angle(slider_slot, &self.constants.back()))
                self.cons_list.push_back(point_on_line(p1, slider_slot))
                if vp1.has_offset():
                    p2 = &self.slider_bases[b]
                    if vp1.offset():
                        self.constants.push_back(vp1.offset())
                        self.cons_list.push_back(p2p_distance(p2, p1, &self.constants.back()))
                    else:
                        self.cons_list.push_back(point_on_point(p2, p1))
            else:
                # Slider between links
                for name in vp1.links[:1]:
                    vlink = self.vlinks[name]
                    # A base link friend
                    c = vlink.points[0]
                    if c == a:
                        if len(vlink.points) < 2:
                            # If no any friend
                            continue
                        c = vlink.points[1]
                    vp2 = self.vpoints[c]
                    if vp2.is_slot_link(vlink.name):
                        # c is a slider, and it is be connected with slot link
                        p2 = &self.slider_bases[self.sliders[c]]
                    else:
                        # c is a R joint or it is not connected with slot link
                        p2 = &self.points[c]
                    self.slider_lines.push_back([&self.slider_bases[b], p2])
                    self.constants.push_back(_radians(vp1.slope_angle(vp2) - vp1.angle))
                    self.cons_list.push_back(internal_angle(
                        slider_slot,
                        &self.slider_lines.back(),
                        &self.constants.back()
                    ))
                    self.cons_list.push_back(point_on_line(p1, slider_slot))
                    if vp1.has_offset():
                        p2 = &self.slider_bases[b]
                        if vp1.offset():
                            self.constants.push_back(vp1.offset())
                            self.cons_list.push_back(p2p_distance(p2, p1, &self.constants.back()))
                        else:
                            self.cons_list.push_back(point_on_point(p2, p1))
            if vp1.type != VJoint.P:
                continue
            for name in vp1.links[1:]:
                vlink = self.vlinks[name]
                # A base link friend
                c = vlink.points[0]
                if c == a:
                    if len(vlink.points) < 2:
                        # If no any friend
                        continue
                    c = vlink.points[1]
                vp2 = self.vpoints[c]
                if vp2.is_slot_link(vlink.name):
                    # c is a slider, and it is be connected with slot link
                    p2 = &self.slider_bases[self.sliders[c]]
                else:
                    # c is a R joint or it is not connected with slot link
                    p2 = &self.points[c]
                self.slider_lines.push_back([p1, p2])
                self.constants.push_back(_radians(vp1.slope_angle(vp2) - vp1.angle))
                self.cons_list.push_back(internal_angle(
                    slider_slot,
                    &self.slider_lines.back(),
                    &self.constants.back()
                ))
        # Angle constraints
        cdef double angle
        for (b, d), angle in self.inputs.items():
            if b == d:
                continue
            self.handles.push_back([&self.points[b], &self.points[d]])
            self.inputs_angle.push_back(_radians(angle))
            self.cons_list.push_back(line_internal_angle(
                &self.handles.back(),
                &self.inputs_angle.back()
            ))

    cdef bint check_known(self, int i):
        """Check known coordinates."""
        if i not in self.data_dict:
            return False
        coord = self.data_dict[i]
        self.data_values.push_back(coord.x)
        tmp_ptr = &self.data_values.back()
        self.data_values.push_back(coord.y)
        self.points.push_back([tmp_ptr, &self.data_values.back()])
        return True

    cpdef void set_inputs(self, dict inputs):
        """Set the values of `inputs` parameter from original constructor.
        Two groups of `dict` keys must be the same or subset.
        """
        if self.inputs is None or inputs is None:
            raise ValueError(f"do not accept modifications")
        if not self.show_inputs() >= set(inputs):
            raise ValueError(f"format must be {set(self.inputs)}, not {set(inputs)}")
        self.inputs.update(inputs)
        # Set values
        cdef int b, d
        cdef double angle
        cdef double *handle
        cdef clist[double].iterator it = self.inputs_angle.begin()
        for (b, d), angle in self.inputs.items():
            if b == d:
                continue
            handle = de_refer_post_inc(it)
            handle[0] = _radians(angle)

    cpdef void set_data(self, dict data_dict):
        """Set the values of `data_dict` parameter from original constructor.
        Two groups of `dict` keys must be the same or subset.
        """
        if self.data_dict is None or data_dict is None:
            raise ValueError(f"do not accept modifications")
        _sort_pairs(data_dict)
        if not self.show_data() >= set(data_dict):
            raise ValueError(f"format must be {set(self.data_dict)}, not {set(data_dict)}")
        self.data_dict.update(data_dict)
        cdef size_t n = 0
        # Set values
        cdef int i
        cdef double *handle
        cdef VPoint vpoint
        cdef Coord coord
        cdef clist[double].iterator it = self.data_values.begin()
        for i, vpoint in enumerate(self.vpoints):
            if vpoint.grounded():
                if i in self.data_dict:
                    # Known coordinates
                    coord = self.data_dict[i]
                    handle = de_refer_post_inc(it)
                    handle[0] = coord.x
                    handle = de_refer_post_inc(it)
                    handle[0] = coord.y
            if i in self.data_dict:
                # Known coordinates
                coord = self.data_dict[i]
                handle = de_refer_post_inc(it)
                handle[0] = coord.x
                handle = de_refer_post_inc(it)
                handle[0] = coord.y
        cdef int a, b, c, d
        cdef VLink vlink
        for vlink in self.vlinks.values():
            if len(vlink.points) < 2:
                continue
            if vlink.name == VLink.FRAME:
                continue
            a = vlink.points[0]
            b = vlink.points[1]
            if (a not in self.data_dict) or (b not in self.data_dict):
                handle = de_refer_post_inc(it)
                handle[0] = self.data_dict[frozenset({a, b})]
            for c in vlink.points[2:]:
                if c in self.data_dict:
                    # Known coordinate
                    continue
                for d in (a, b):
                    handle = de_refer_post_inc(it)
                    handle[0] = self.data_dict[frozenset({c, d})]

    cpdef list solve(self):
        """Solve the conditions and return the result, raise ValueError if
        not succeeded.
        The joint position will returned by its index correspondingly.

        + Revolute joints: Tuple[float, float]
        + Slider joints: Tuple[Tuple[float, float], Tuple[float, float]]
        """
        # Pointer of parameters
        cdef size_t params_count = <int>self.params.size()
        cdef double **params_ptr = <double **>PyMem_Malloc(sizeof(double *) * params_count)
        cdef clist[double].iterator it = self.params.begin()
        cdef size_t i
        for i in range(params_count):
            params_ptr[i] = de_refer_post_inc(it)
        # Pointer of constraints
        cdef size_t cons_count = <int>self.cons_list.size()
        cdef Constraint *cons = <Constraint *>PyMem_Malloc(sizeof(Constraint) * cons_count)
        i = 0
        cdef Constraint con
        for con in self.cons_list:
            cons[i] = con
            i += 1
        # Solve
        cdef bint flag = solve(params_ptr, params_count, cons, cons_count, False)
        cdef VPoint vp
        cdef Coord c
        cdef Point p1, p2
        if flag:
            solved_points = []
            for i, vp in enumerate(self.vpoints):
                if i in self.data_dict or vp.no_link():
                    if vp.no_link():
                        c = Coord.__new__(Coord, vp.c[0, 0], vp.c[0, 1])
                    else:
                        c = self.data_dict[i]
                    if vp.type == VJoint.R:
                        solved_points.append((c.x, c.y))
                    else:
                        solved_points.append(((c.x, c.y), (c.x, c.y)))
                    continue
                p1 = self.points[i]
                if vp.type == VJoint.R:
                    solved_points.append((p1.x[0], p1.y[0]))
                else:
                    p2 = self.slider_bases[self.sliders[i]]
                    solved_points.append(((p2.x[0], p2.y[0]),
                                          (p1.x[0], p1.y[0])))
        PyMem_Free(params_ptr)
        PyMem_Free(cons)
        if flag:
            return solved_points
        else:
            raise ValueError("no valid solutions were found from initialed values")



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Triangular expressions.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from libc.math cimport sin, cos, M_PI
from libcpp.map cimport map
from numpy import zeros, float64 as f64
from .expression cimport VJoint, VPoint, VLink
from .tinycadlib cimport (
    PXY, PPP, PLA, PLAP, PLLP, PLPP, PALP, P_LABEL, L_LABEL, I_LABEL,
    A_LABEL, S_LABEL
)

ctypedef vector[Sym] Inputs
ctypedef map[int, bint] Status


cdef str symbol_str(Sym p):
    """The match pattern of the symbols."""
    if p.first == P_LABEL:
        return f"P{p.second}"
    elif p.first == L_LABEL:
        return f"L{p.second}"
    elif p.first == I_LABEL:
        return f"I{p.second}"
    elif p.first == A_LABEL:
        return f"A{p.second}"
    elif p.first == S_LABEL:
        return f"S{p.second}"
    else:
        return ""


cdef class EStack:
    """Triangle solution stack, generated from [`t_config`](#t_config).
    It is pointless to call the constructor.
    """

    cdef void add_pxy(self, Sym c1, Sym v1, Sym v2, Sym t) nogil:
        cdef Expr e
        e.func = PXY
        e.c1 = c1
        e.v1 = v1
        e.v2 = v2
        e.target = t
        e.op = False
        self.stack.push_back(e)

    cdef void add_ppp(self, Sym c1, Sym c2, Sym c3, Sym t) nogil:
        cdef Expr e
        e.func = PPP
        e.c1 = c1
        e.c2 = c2
        e.c3 = c3
        e.target = t
        self.stack.push_back(e)

    cdef void add_pla(self, Sym c1, Sym v1, Sym v2, Sym t) nogil:
        cdef Expr e
        e.func = PLA
        e.c1 = c1
        e.v1 = v1
        e.v2 = v2
        e.target = t
        e.op = False
        self.stack.push_back(e)

    cdef void add_plap(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t) nogil:
        cdef Expr e
        e.func = PLAP
        e.c1 = c1
        e.v1 = v1
        e.v2 = v2
        e.c2 = c2
        e.target = t
        e.op = False
        self.stack.push_back(e)

    cdef void add_pllp(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t) nogil:
        cdef Expr e
        e.func = PLLP
        e.c1 = c1
        e.v1 = v1
        e.v2 = v2
        e.c2 = c2
        e.target = t
        e.op = False
        self.stack.push_back(e)

    cdef void add_plpp(self, Sym c1, Sym v1, Sym c2, Sym c3, Sym t, bint op) nogil:
        cdef Expr e
        e.func = PLPP
        e.c1 = c1
        e.v1 = v1
        e.c2 = c2
        e.c3 = c3
        e.target = t
        e.op = op
        self.stack.push_back(e)

    cdef void add_palp(self, Sym c1, Sym v1, Sym v2, Sym c2, Sym t, bint op) nogil:
        cdef Expr e
        e.func = PALP
        e.c1 = c1
        e.v1 = v1
        e.v2 = v2
        e.c2 = c2
        e.target = t
        e.op = op
        self.stack.push_back(e)

    cpdef list as_list(self):
        """Copy the dataset as list object."""
        stack = []
        cdef Expr expr
        for expr in self.stack:
            if expr.func == PXY:
                stack.append((
                    "PXY",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.v2),
                    symbol_str(expr.target),
                ))
            elif expr.func == PPP:
                stack.append((
                    "PPP",
                    symbol_str(expr.c1),
                    symbol_str(expr.c2),
                    symbol_str(expr.c3),
                    symbol_str(expr.target),
                ))
            elif expr.func == PLA:
                stack.append((
                    "PLA",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.v2),
                    symbol_str(expr.target),
                ))
            elif expr.func == PLAP:
                stack.append((
                    "PLAP",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.v2),
                    symbol_str(expr.c2),
                    symbol_str(expr.target),
                ))
            elif expr.func == PLLP:
                stack.append((
                    "PLLP",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.v2),
                    symbol_str(expr.c2),
                    symbol_str(expr.target),
                ))
            elif expr.func == PLPP:
                stack.append((
                    "PLPP",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.c2),
                    symbol_str(expr.c3),
                    symbol_str(expr.target),
                ))
            elif expr.func == PALP:
                stack.append((
                    "PALP",
                    symbol_str(expr.c1),
                    symbol_str(expr.v1),
                    symbol_str(expr.v2),
                    symbol_str(expr.c2),
                    symbol_str(expr.target),
                ))
        return stack

    def __repr__(self):
        return f"{type(self).__name__}({self.as_list()})"


cdef bint _is_all_lock(Status &status):
    """Test is all status done."""
    cdef bint n_status
    for _, n_status in status:
        if not n_status:
            return False
    return True


cdef bint _is_parallel(VPoint p1, VPoint p2, VPoint p3, VPoint p4):
    """Check parallel four bar loop."""
    return (abs(p1.distance(p3) - p2.distance(p4)) < 1e-12
            and abs(p2.distance(p3) - p1.distance(p4)) < 1e-12)


cdef bint _clockwise(double[:] c1, double[:] c2, double[:] c3):
    """Check the orientation of three points."""
    cdef double val = (c2[1] - c1[1]) * (c3[0] - c2[0]) - (c2[0] - c1[0]) * (c3[1] - c2[1])
    return val == 0 or val > 0


cdef (bint, int, int) _get_reliable_friends(
    int node,
    VPoint vpoint,
    object vlinks,
    Status &status
):
    """Return two "has been solved" nodes on the same link."""
    cdef int fa = -1
    cdef int fb = -1
    cdef int f
    for link in vpoint.links:
        if len(vlinks[link]) < 2:
            continue
        for f in vlinks[link]:
            if status[f] and f != node:
                if fa == -1:
                    fa = f
                elif fb == -1:
                    fb = f
                    return True, fa, fb
    return False, -1, -1


cdef int _get_intersection(
    int node1,
    int node2,
    object vpoints,
    object vlinks,
    Status &status
):
    """Get a configured node between two nodes."""
    cdef int node
    cdef VPoint vp
    for node, vp in enumerate(vpoints):
        if (set(vp.links) & set((<VPoint>vpoints[node1]).links)
            and set(vp.links) & set((<VPoint> vpoints[node2]).links)
            and status[node]
        ):
            return node
    return -1


cdef int _get_not_base_friend(VPoint vpoint, object vlinks, Status &status):
    """Get a configured node from other links."""
    if (vpoint.pin_grounded()
        or not vpoint.grounded()
        or vpoint.has_offset()
        or len(vpoint.links) < 2
    ):
        return -1
    cdef int f
    for f in vlinks[vpoint.links[1]]:
        if status[f]:
            return f
    return -1


cdef (bint, int, int) _get_base_friend(VPoint vpoint, object vlinks):
    """Get two configured nodes on the same links."""
    if len(vpoint.links) < 1:
        return False, -1, -1
    cdef int fa = -1
    cdef int f
    for f in vlinks[vpoint.links[0]]:
        if fa == -1:
            fa = f
        else:
            return True, fa, f
    return False, -1, -1


cdef int _get_input_base(int node, Inputs &inputs):
    """Get the base node for input pairs."""
    cdef int base, node_
    for base, node_ in inputs:
        if node == node_:
            return base
    return -1


cpdef EStack t_config(
    object vpoints_,
    object inputs_,
    object status_ = None
):
    """Generate the Triangle solution stack by mechanism expression `vpoints_`.

    The argument `inputs` is a list of input pairs.
    The argument `status` will track the configuration of each point,
    which is optional.
    """
    # For VPoint list:
    # + vpoints_: [vpoint0, vpoint1, ...]
    # + inputs: [(p0, p1), (p0, p2), ...]
    # + status: Dict[int, bint]
    # vpoints will make a copy that we don't want to modified itself
    cdef bint has_input = True
    cdef bint has_status = True
    if inputs_ is None or not inputs_:
        has_input = False
    if status_ is None:
        has_status = False
    cdef EStack exprs = EStack.__new__(EStack)
    if not vpoints_ or not has_input:
        return exprs

    vpoints = list(vpoints_)
    cdef Inputs inputs
    cdef Status status

    cdef bint ok
    cdef int node, base
    if has_input:
        for node, base in inputs_:
            inputs.push_back(Sym(node, base))
    if has_status:
        for node, ok in status_.items():
            status[node] = ok
    # First, we create a "VLinks" that can help us to
    # find a relationship just like adjacency matrix
    cdef VPoint vp1, vp2, vp3
    vlinks = {}
    for node, vp1 in enumerate(vpoints):
        status[node] = False
        if vp1.links:
            for link in vp1.links:
                # Connect on the ground and it is not a slider
                if link == VLink.FRAME and vp1.type == VJoint.R:
                    status[node] = True
                # Add as vlink
                if link not in vlinks:
                    vlinks[link] = {node}
                else:
                    vlinks[link].add(node)
        else:
            status[node] = True
    # Replace the P joints and their friends with RP joint
    # DOF must be same after properties changed
    for base, vp1 in enumerate(vpoints):
        if vp1.type != VJoint.P or not vp1.grounded():
            continue
        for link in vp1.links[1:]:
            for node in vlinks[link]:
                vp2 = vpoints[node]
                if node == base or vp2.type != VJoint.R:
                    continue
                vpoints[node] = VPoint.c_slider_joint([vp1.links[0]] + [
                    link_ for link_ in vp2.links
                    if link_ not in vp1.links
                ], VJoint.RP, vp1.angle, vp2.x, vp2.y)
    # Add positions parameters
    cdef double[:, :] pos = zeros((len(vpoints), 2), dtype=f64)
    for base, vp1 in enumerate(vpoints):
        node = 0 if vp1.type == VJoint.R else 1
        pos[base, 0] = vp1.c[node, 0]
        pos[base, 1] = vp1.c[node, 1]
    cdef int link_symbol = 0
    cdef int input_symbol = 0
    cdef int angle_symbol = 0
    # Input joints (R) that was connect with ground
    for base, node in inputs:
        if status[base]:
            exprs.add_pla(
                [P_LABEL, base],
                [L_LABEL, link_symbol],
                [I_LABEL, input_symbol],
                [P_LABEL, node]
            )
            status[node] = True
            link_symbol += 1
            input_symbol += 1
    # Now let we search around all the points, until we find the solutions
    # that we could
    input_targets = {node for _, node in inputs}
    node = 0
    cdef int skip_times = 0
    cdef int around = len(status)
    cdef int fa, fb, fc, fd
    cdef double angle
    cdef double[:] tmp = zeros(2, dtype=f64)
    # Friend iterator
    while not _is_all_lock(status):
        if status.find(node) == status.end():
            node = 0
            continue
        # Check and break the loop if it's re-scan again
        if skip_times >= around:
            break
        if status[node]:
            node += 1
            skip_times += 1
            continue
        vp1 = vpoints[node]
        if vp1.type == VJoint.R:
            # R joint
            # + Is input node?
            # + Normal revolute joint
            if node in input_targets:
                base = _get_input_base(node, inputs)
                if status[base]:
                    exprs.add_pla(
                        Sym(P_LABEL, base),
                        Sym(L_LABEL, link_symbol),
                        Sym(I_LABEL, input_symbol),
                        Sym(P_LABEL, node)
                    )
                    status[node] = True
                    link_symbol += 1
                    input_symbol += 1
                else:
                    skip_times += 1
            else:
                ok, fa, fb = _get_reliable_friends(node, vp1, vlinks, status)
                if not ok:
                    skip_times += 1
                else:
                    if not _clockwise(pos[fa], pos[node], pos[fb]):
                        fa, fb = fb, fa
                    vp2 = vpoints[fa]
                    vp3 = vpoints[fb]
                    if vp2.same_link(vp3) and not (vp2.grounded() and vp3.grounded()):
                        exprs.add_plap(
                            Sym(P_LABEL, fa),
                            Sym(L_LABEL, link_symbol),
                            Sym(A_LABEL, angle_symbol),
                            Sym(P_LABEL, fb),
                            Sym(P_LABEL, node)
                        )
                        link_symbol += 1
                        angle_symbol += 1
                    else:
                        # Decide when to solve parallel linkage
                        fc = _get_intersection(fa, fb, vpoints, vlinks, status)
                        if (fc != -1 and _is_parallel(
                            vpoints[fa], vpoints[fb], vpoints[fc], vpoints[node]
                        )):
                            exprs.add_ppp(
                                Sym(P_LABEL, fc),
                                Sym(P_LABEL, fa),
                                Sym(P_LABEL, fb),
                                Sym(P_LABEL, node)
                            )
                        else:
                            exprs.add_pllp(
                                Sym(P_LABEL, fa),
                                Sym(L_LABEL, link_symbol),
                                Sym(L_LABEL, link_symbol + 1),
                                Sym(P_LABEL, fb),
                                Sym(P_LABEL, node)
                            )
                            link_symbol += 2
                    status[node] = True
                    skip_times = 0
        elif vp1.type == VJoint.P:
            # Need to solve P joint itself here (only grounded)
            fa = _get_not_base_friend(vp1, vlinks, status)
            if fa == -1:
                skip_times += 1
            else:
                exprs.add_pxy(
                    Sym(P_LABEL, fa),
                    Sym(L_LABEL, link_symbol),
                    Sym(L_LABEL, link_symbol + 1),
                    Sym(P_LABEL, node)
                )
                status[node] = True
                link_symbol += 2
                # Solution for all friends
                for link in vp1.links[1:]:
                    for fb in vlinks[link]:
                        if status[fb]:
                            continue
                        exprs.add_pxy(
                            Sym(P_LABEL, node),
                            Sym(L_LABEL, link_symbol),
                            Sym(L_LABEL, link_symbol + 1),
                            Sym(P_LABEL, fb)
                        )
                        status[fb] = True
                        link_symbol += 2
                skip_times = 0
        elif vp1.type == VJoint.RP:
            # Copy as 'fc'
            fc = node
            # 'S' point
            tmp[:] = pos[node, :]
            angle = vp1.angle / 180 * M_PI
            tmp[0] += cos(angle)
            tmp[1] += sin(angle)
            fa = _get_not_base_friend(vp1, vlinks, status)
            ok, fb, fd = _get_base_friend(vp1, vlinks)
            if fa == -1 or not ok:
                skip_times += 1
            else:
                # Slot is not grounded
                if not vp1.grounded():
                    if not _clockwise(pos[fb], tmp, pos[fd]):
                        fb, fd = fd, fb
                    exprs.add_pllp(
                        Sym(P_LABEL, fb),
                        Sym(L_LABEL, link_symbol),
                        Sym(L_LABEL, link_symbol + 1),
                        Sym(P_LABEL, fd),
                        Sym(P_LABEL, node)
                    )
                    link_symbol += 2
                # PLPP
                # [PLLP]
                # Set 'S' (slider) point to define second point of slider
                # + A 'friend' from base link
                # + Get distance from me and friend
                # [PLPP]
                # Re-define coordinate of target point by self and 'S' point
                # + A 'friend' from other link
                # + Solve
                if not _clockwise(pos[fb], tmp, pos[fc]):
                    fb, fc = fc, fb
                exprs.add_pllp(
                    Sym(P_LABEL, fb),
                    Sym(L_LABEL, link_symbol),
                    Sym(L_LABEL, link_symbol + 1),
                    Sym(P_LABEL, fc),
                    Sym(S_LABEL, node)
                )
                # Two conditions
                exprs.add_plpp(
                    Sym(P_LABEL, fa),
                    Sym(L_LABEL, link_symbol + 2),
                    Sym(P_LABEL, node),
                    Sym(S_LABEL, node),
                    Sym(P_LABEL, node),
                    (pos[fa, 0] - pos[node, 0] > 0) != (vp1.angle > 90)
                )
                status[node] = True
                link_symbol += 3
                skip_times = 0
        node += 1
    if has_status:
        for node, ok in status:
            status_[node] = ok
    return exprs



# -*- coding: utf-8 -*-

from typing import Callable, Optional
from .utility import Algorithm, ObjFunc, FVal
from .config_types import FASetting

class FA(Algorithm):

    def __init__(
        self,
        func: ObjFunc[FVal],
        settings: FASetting,
        progress_fun: Optional[Callable[[int, str], None]] = None,
        interrupt_fun: Optional[Callable[[], bool]] = None
    ):
        """The format of argument `settings`:

        + `pop_num`: Population
            + type: int
            + default: 80
        + `alpha`: Alpha factor
            + type: float (0.~1.)
            + default: 0.01
        + `beta_min`: Minimal attraction
            + type: float (0.~1.)
            + default: 0.2
        + `beta0`: Attraction rate
            + type: float (0.~1.)
            + default: 1.
        + `gamma`: Gamma rate
            + type: float (0.~1.)
            + default: 1.
        + `max_gen` or `min_fit` or `max_time`: Limitation of termination
            + type: int / float / float
            + default: Raise `ValueError`
        + `report`: Report per generation
            + type: int
            + default: 10

        Others arguments are same as [`Differential.__init__()`](#differential9595init__).
        """
        ...



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Sharing position analysis function.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

from libcpp.pair cimport pair
from libcpp.map cimport map
from libcpp.vector cimport vector
from .topo_config cimport EStack

cdef extern from "tinycadlib/solver.h" nogil:
    struct SwappablePair:
        int first, second

    enum Label:
        P_LABEL
        L_LABEL
        I_LABEL
        A_LABEL
        S_LABEL

    enum Func:
        PXY
        PPP
        PLA
        PLAP
        PLLP
        PLPP
        PALP

    struct CCoord:
        double x, y

    CCoord cpxy(CCoord c1, double x, double y)
    CCoord cppp(CCoord c1, CCoord c2, CCoord c3)
    CCoord cplap(CCoord c1, double d0, double a0, CCoord c2, bint inverse)
    CCoord cpllp(CCoord c1, double d0, double d1, CCoord c2, bint inverse)
    CCoord cplpp(CCoord c1, double d0, CCoord c2, CCoord c3, bint inverse)
    CCoord cpalp(CCoord c1, double a0, double d0, CCoord c2, bint inverse)

    ctypedef pair[int, int] Sym

    struct Expr:
        bint op
        Func func
        Sym v1, v2, c1, c2, c3, target

    cppclass ExprSolver:
        map[Sym, CCoord] joint_pos

        ExprSolver()
        ExprSolver(vector[Expr] stack, map[Sym, CCoord] joint_pos,
                   map[Sym, double] param)
        bint solve()

cdef bint preprocessing(EStack exprs, object vpoints, object inputs,
                        map[Sym, CCoord] & joint_pos,
                        map[SwappablePair, double] & link_len,
                        map[Sym, double] & param)
cpdef list expr_solving(EStack exprs, object vpoints, object inputs=*)
cdef (bint, map[Sym, CCoord]) quick_solve(vector[Expr] stack,
                                          map[Sym, CCoord] joint_pos,
                                          map[Sym, double] param) nogil
cdef double[:, :, :] c_uniform_path(double[:, :] dimension, int n) nogil
cpdef object uniform_expr(double[:] v)



# -*- coding: utf-8 -*-

from typing import TypedDict
from .de import Strategy


class Setting(TypedDict, total=False):
    pop_num: int
    max_gen: int
    min_fit: float
    max_time: float
    slow_down: float
    report: int
    parallel: bool


class DESetting(Setting):
    strategy: Strategy
    f: float
    cr: float


class RGASetting(Setting):
    cross: float
    mutate: float
    win: float
    delta: float


class PSOSetting(Setting):
    cognition: float
    social: float
    velocity: float


class FASetting(Setting):
    alpha: float
    beta_min: float
    beta0: float
    gamma: float


class TOBLSetting(Setting):
    pass



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Teaching Learning Based Optimization

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libc.math cimport round
from numpy import zeros, float64 as f64
from .utility cimport uint, rand_v, rand_i, ObjFunc, Algorithm


@cython.final
cdef class TLBO(Algorithm):
    """The implementation of Teaching Learning Based Optimization."""
    cdef double[:] tmp

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        self.tmp = zeros(self.dim, dtype=f64)

    cdef inline void bounding(self, uint s) nogil:
        if self.tmp[s] < self.func.lb[s]:
            self.tmp[s] = self.func.lb[s]
        elif self.tmp[s] > self.func.ub[s]:
            self.tmp[s] = self.func.ub[s]

    cdef inline void register(self, uint i) nogil:
        cdef double f_new = self.func.fitness(self.tmp)
        if f_new < self.fitness[i]:
            self.pool[i, :] = self.tmp
            self.fitness[i] = f_new
        if self.fitness[i] < self.best_f:
            self.set_best(i)

    cdef inline void teaching(self, uint i) nogil:
        """Teaching phase. The last best is the teacher."""
        cdef double tf = round(1 + rand_v())
        cdef uint s, j
        cdef double mean
        for s in range(self.dim):
            mean = 0
            for j in range(self.pop_num):
                mean += self.pool[j, s]
            mean /= self.dim
            self.tmp[s] = self.pool[i, s] + rand_v(1, self.dim) * (
                self.best[s] - tf * mean)
            self.bounding(s)
        self.register(i)

    cdef inline void learning(self, uint i) nogil:
        """Learning phase."""
        cdef uint j = rand_i(self.pop_num - 1)
        if j >= i:
            j += 1
        cdef uint s
        cdef double diff
        for s in range(self.dim):
            if self.fitness[j] < self.fitness[i]:
                diff = self.pool[i, s] - self.pool[j, s]
            else:
                diff = self.pool[j, s] - self.pool[i, s]
            self.tmp[s] = self.pool[i, s] + diff * rand_v(1, self.dim)
            self.bounding(s)
        self.register(i)

    cdef inline void generation(self) nogil:
        """The process of each generation."""
        cdef uint i
        for i in range(self.pop_num):
            self.teaching(i)
            self.learning(i)



# -*- coding: utf-8 -*-

from typing import Tuple, List, Sequence, Dict, Optional
from .expression import VPoint

class EStack:
    well_done: bool

    def as_list(self) -> List[Tuple[str, ...]]:
        ...

def t_config(
    vpoints: Sequence[VPoint],
    inputs: Sequence[Tuple[int, int]],
    status: Optional[Dict[int, bool]] = None
) -> EStack:
    ...



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""PMKS symbolics.

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libc.math cimport M_PI, atan2, hypot
from cpython.object cimport Py_EQ, Py_NE
from numpy import array, zeros, float64 as f64


cdef double distance(double x1, double y1, double x2, double y2) nogil:
    """Distance of two coordinates."""
    return hypot(x2 - x1, y2 - y1)


cdef double slope_angle(double x1, double y1, double x2, double y2) nogil:
    """Slope angle of two coordinates."""
    return atan2(y1 - y2, x1 - x2)


cpdef list get_vlinks(object vpoints):
    """Get VLinks from a list of VPoint `vpoints`."""
    links = {}
    cdef int i
    cdef VPoint vpoint
    for i, vpoint in enumerate(vpoints):
        for name in vpoint.links:
            if name not in links:
                links[name] = {i}
            else:
                links[name].add(i)
    vlinks = []
    for name, points in links.items():
        vlinks.append(VLink(name, "", points))
    return vlinks


@cython.final
cdef class Coord:
    """A data class used to store coordinates."""

    def __cinit__(self, double x, double y):
        self.x = x
        self.y = y

    cpdef double distance(self, Coord p):
        """Return the distance between two coordinates."""
        return distance(self.x, self.y, p.x, p.y)

    cpdef double slope_angle(self, Coord p):
        """Slope angle of two coordinates."""
        return slope_angle(self.x, self.y, p.x, p.y)

    cpdef bint is_nan(self):
        """Return true if the coordinate value is not a number."""
        return self.x != self.x

    def __repr__(self):
        return f"Coord({self.x:.02f}, {self.y:.02f})"


@cython.final
cdef class VPoint:
    """Mechanism expression class."""
    HOLDER = VPoint([], VJoint.R, 0., "", 0., 0.)

    def __cinit__(
        self,
        object links,
        VJoint j_type,
        double angle,
        str color_str,
        double x,
        double y,
        object color_func=None
    ):
        self.set_links(links)
        self.type = j_type
        self.type_str = ('R', 'P', 'RP')[j_type]
        self.angle = angle
        self.color_str = color_str
        if color_func is None:
            self.color = None
        else:
            self.color = color_func(color_str)
        self.x = x
        self.y = y
        self.c = zeros((2, 2), dtype=f64)
        if self.type in {VJoint.P, VJoint.RP}:
            # Slider current coordinates
            # [0]: Current node on slot
            # [1]: Pin
            self.c[0, 0] = self.c[1, 0] = self.x
            self.c[0, 1] = self.c[1, 1] = self.y
        else:
            self.c[0, 0] = self.x
            self.c[0, 1] = self.y
        self.__has_offset = False
        self.__offset = 0

    @staticmethod
    def r_joint(links, x, y):
        """A fast constructor of revolute joints."""
        return VPoint.c_r_joint(links, x, y)

    @staticmethod
    cdef VPoint c_r_joint(object links, double x, double y):
        return VPoint.__new__(VPoint, links, VJoint.R, 0., '', x, y)

    @staticmethod
    def slider_joint(links, type_int, angle, x, y):
        """A fast constructor of slider joints."""
        return VPoint.c_slider_joint(links, type_int, angle, x, y)

    @staticmethod
    cdef VPoint c_slider_joint(object links, VJoint type_int, double angle, double x, double y):
        return VPoint.__new__(VPoint, links, type_int, angle, '', x, y)

    cpdef VPoint copy(self):
        """The copy method of the VPoint object."""
        return self.__copy__()

    @property
    def sx(self):
        """X value of slot coordinate."""
        return self.c[0, 0]

    @property
    def sy(self):
        """Y value of slot coordinate."""
        return self.c[0, 1]

    @property
    def cx(self):
        """X value of current coordinate.
        If it's slider, the pin coordinate will be returned.
        """
        if self.type == VJoint.R:
            return self.c[0, 0]
        else:
            return self.c[1, 0]

    @property
    def cy(self):
        """Y value of current coordinate.
        If it's slider, the pin coordinate will be returned.
        """
        if self.type == VJoint.R:
            return self.c[0, 1]
        else:
            return self.c[1, 1]

    cpdef void set_links(self, object links) except *:
        """The update function of links attribute."""
        self.links = tuple([s for s in links if s])

    cpdef void replace_link(self, str link1, str link2) except *:
        """Replace the value in links attribute."""
        self.set_links([link2 if link == link1 else link for link in self.links])

    cpdef void move(self, object c1, object c2 = None) except *:
        """The update function of current coordinate(s).
        The 2nd placement is the pin coordinate of slider joints.

        If there is only one argument for a slider joint,
        the slot and pin coordinates will be set to the same position.
        """
        cdef double x, y
        x, y = c1
        self.c[0, 0] = x
        self.c[0, 1] = y
        if self.type in {VJoint.P, VJoint.RP}:
            if c2 is not None:
                x, y = c2
            self.c[1, 0] = x
            self.c[1, 1] = y

    cpdef void locate(self, double x, double y) except *:
        """The update function of original coordinate."""
        self.x = x
        self.y = y
        self.c[0, 0] = x
        self.c[0, 1] = y

    cpdef void rotate(self, double angle):
        """The update function of angle attribute."""
        self.angle = angle % 180

    cpdef void set_offset(self, double offset):
        """The update function of slider offset.
        It will also enable offset value after called.
        """
        self.__has_offset = True
        self.__offset = offset

    cpdef void disable_offset(self):
        """Disable offset setting of the joint."""
        self.__has_offset = False

    cpdef bint is_slider(self):
        """Return true for slider type."""
        return self.type in {VJoint.P, VJoint.RP}

    cpdef double distance(self, VPoint p):
        """Return the distance between two VPoint objects."""
        on_links = tuple(set(self.links) & set(p.links))
        cdef double m_x = 0
        cdef double m_y = 0
        cdef double p_x = 0
        cdef double p_y = 0
        if on_links:
            if self.type == VJoint.R or self.links[0] == on_links[0]:
                # self is R joint or at base link
                m_x = self.c[0, 0]
                m_y = self.c[0, 1]
            else:
                # At pin joint
                m_x = self.c[1, 0]
                m_y = self.c[1, 1]
            if p.type == VJoint.R or p.links[0] == on_links[0]:
                # p is R joint or at base link
                p_x = p.c[0, 0]
                p_y = p.c[0, 1]
            else:
                # At pin joint
                p_x = p.c[1, 0]
                p_y = p.c[1, 1]
        else:
            m_x = self.c[0, 0]
            m_y = self.c[0, 1]
            p_x = p.c[0, 0]
            p_y = p.c[0, 1]
        return hypot(p_x - m_x, p_y - m_y)

    cpdef bint has_offset(self):
        """Return true if the offset setting is enabled."""
        return self.__has_offset

    cpdef double offset(self):
        """Return the offset constraint value of the joint."""
        return self.__offset

    cpdef double true_offset(self):
        """Return the current offset value of the joint."""
        return hypot(self.c[1, 0] - self.c[0, 0], self.c[1, 1] - self.c[0, 1])

    cpdef double slope_angle(self, VPoint p, int num1 = 2, int num2 = 2):
        """Return the value `hypot(p_x - m_x, p_y - m_y)`,
        where `m_x`, `m_y` is the value of the joint,
        and `p_x`, `p_y` is the value of `p`.

        The option `num1` and `num2` is the position of current coordinate
        attribute.
        """
        cdef double x1, y1, x2, y2
        if num1 > 1:
            x2, y2 = self.x, self.y
        else:
            x2 = self.c[num2, 0]
            y2 = self.c[num2, 1]
        if num2 > 1:
            x1, y1 = p.x, p.y
        else:
            x1 = p.c[num2, 0]
            y1 = p.c[num2, 1]
        return slope_angle(x1, y1, x2, y2) / M_PI * 180

    cpdef Coord link_pos(self, str link):
        """Return the position for the vlink."""
        cdef size_t ind
        if self.type == VJoint.R or self.is_slot_link(link):
            ind = 0
        else:
            ind = 1
        return self.to_coord(ind)

    cpdef bint grounded(self):
        """Return true if the joint pin is connected to ground link."""
        if self.type == VJoint.R:
            return VLink.FRAME in self.links
        elif self.type in {VJoint.P, VJoint.RP}:
            if self.links:
                return self.is_slot_link(VLink.FRAME)
            else:
                return False

    cpdef bint pin_grounded(self):
        """Return true if the point is at the same link."""
        return VLink.FRAME in self.links[1:]

    cpdef bint same_link(self, VPoint p):
        """Return true if the point is at the same link."""
        return bool(set(self.links) & set(p.links))

    cpdef bint no_link(self):
        """Return true if there is no any link in links attribute."""
        return not self.links

    cpdef bint is_slot_link(self, str link):
        """Return true if the slot is on the link `link_name`."""
        if self.type == VJoint.R:
            return False
        if self.links:
            return link == self.links[0]
        else:
            return False

    cpdef str expr(self):
        """Return the literal mechanism expression of the joint."""
        if self.type != VJoint.R:
            type_text = f"{self.type_str}, A[{self.angle}]"
        else:
            type_text = 'R'
        if self.color_str:
            color = f", color[{self.color_str}]"
        else:
            color = ""
        links_text = ", ".join([name for name in self.links])
        x_text = f"{self.x:.4f}".rstrip('0').rstrip('.')
        y_text = f"{self.y:.4f}".rstrip('0').rstrip('.')
        return f"J[{type_text}{color}, P[{x_text}, {y_text}], L[{links_text}]]"

    cpdef Coord to_coord(self, size_t ind):
        """Obtain coordinate by Coord object."""
        return Coord.__new__(Coord, self.c[ind, 0], self.c[ind, 1])

    def __copy__(self):
        cdef VPoint vpoint = VPoint.__new__(
            VPoint,
            self.links,
            self.type,
            self.angle,
            self.color_str,
            self.x,
            self.y
        )
        vpoint.c[:] = self.c[:]
        return vpoint

    def __richcmp__(self, rhs, int op):
        cdef VPoint other = rhs
        cdef bint different = (
            self.links != other.links or
            self.c[0, 0] != other.c[0, 0] or
            self.c[0, 1] != other.c[0, 1] or
            self.c[1, 1] != other.c[1, 1] or
            self.c[1, 1] != other.c[1, 1] or
            self.type != other.type or
            self.x != other.x or
            self.y != other.y or
            self.angle != other.angle
        )
        if op == Py_EQ:
            return not different
        elif op == Py_NE:
            return different
        else:
            raise TypeError(
                f"'{op}' not support between instances of "
                f"{type(self)} and {type(other)}"
            )

    def __getitem__(self, ind):
        cdef int i = ind
        if self.type == VJoint.R:
            return self.c[0, i]
        else:
            return self.c[1, i]

    def __repr__(self):
        return f"VPoint({self.links}, {int(self.type)}, {self.angle}, {array(self.c).tolist()})"


@cython.final
cdef class VLink:
    """Mechanism expression class in link's view."""
    HOLDER = VLink("", "", [])
    FRAME = 'ground'

    def __cinit__(
        self,
        str name,
        str color_str,
        object points,
        object color_func=None
    ):
        self.name = name
        self.color_str = color_str
        if color_func is None:
            self.color = None
        else:
            self.color = color_func(color_str)
        self.points = array(list(points), dtype=int)

    cpdef void set_points(self, object points) except *:
        """The update function of points attribute."""
        self.points = array(list(points), dtype=int)

    cpdef Coord[:] points_pos(self, object vpoints) except *:
        """Get link positions from a VPoint list."""
        coords = []
        cdef int i
        cdef Coord c
        cdef VPoint vpoint
        for i in self.points:
            vpoint = vpoints[i]
            coords.append(vpoint.link_pos(self.name))
        return array(coords, dtype=object)

    def __contains__(self, point: int):
        return point in self.points

    def __repr__(self):
        return f"VLink('{self.name}', {tuple(self.points)}, color_qt)"



# -*- coding: utf-8 -*-
# cython: language_level=3

"""Fast planar linkage synthesis. (defect allowable)

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cdef void roll(double[:, :] path, int ind) nogil



# -*- coding: utf-8 -*-

from typing import Iterable, Tuple
from numpy import double, ndarray
from pyslvs.metaheuristics import ObjFunc
from .utility import NConfig

_Coord = Tuple[float, float]

def norm_pca(path: Iterable[_Coord]) -> ndarray:
    ...


class NPlanar(ObjFunc[str]):
    def __init__(self, mech: NConfig):
        """The constructor of objective object.

        Options:
        + `target`: The target path.
        """

    def fitness(self, v: ndarray) -> double:
        ...

    def result(self, v: ndarray) -> str:
        ...



# -*- coding: utf-8 -*-

from os import walk
from os.path import sep, join as pth_join
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext
from platform import system

root_path = 'pyslvs'
bfgs_path = pth_join(root_path, 'bfgs_solver')
tinycadlib_path = pth_join(root_path, 'tinycadlib')
macros = [('_USE_MATH_DEFINES', None)]
compile_args_msvc = ['/O2', '/std:c++17']
compile_args = ['-Wno-cpp', '-std=c++17']
link_args = []
link_args_msvc = []
link_args_static = [
    '-static-libgcc',
    '-static-libstdc++',
    '-Wl,-Bstatic,--whole-archive',
    '-lwinpthread',
    '-Wl,--no-whole-archive',
]
if system() == 'Windows':
    # Disable format warning
    compile_args.append('-Wno-format')
# Disable NumPy warning
macros.append(('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'))
# Special modules mixed with C++ (language must be unified)
ext_modules = [
    Extension(root_path.replace(sep, '.') + '.bfgs', [
        pth_join(root_path, 'bfgs.pyx'),
        pth_join(bfgs_path, 'constraints.cpp'),
        pth_join(bfgs_path, 'solve.cpp'),
        pth_join(bfgs_path, 'calc.cpp'),
    ], language="c++", include_dirs=[bfgs_path]),
    Extension(root_path.replace(sep, '.') + '.tinycadlib', [
        pth_join(root_path, 'tinycadlib.pyx'),
        pth_join(tinycadlib_path, 'solver.cpp'),
    ], language="c++", include_dirs=[tinycadlib_path]),
]
# To find other sources recursively
for root, _, files in walk(root_path):
    for source in files:
        if not source.endswith('.pyx'):
            continue
        if root == root_path and source in {'bfgs.pyx', 'tinycadlib.pyx'}:
            continue
        f_name = pth_join(root, source)
        ext_modules.append(Extension(
            f_name.replace(sep, '.').rsplit('.', maxsplit=1)[0],  # Name
            [f_name],
            language="c++", include_dirs=list({root_path, root})
        ))
for ext in ext_modules:
    ext.cython_directives = {'binding': True, 'cdivision': True}


class Build(build_ext):

    def build_extensions(self):
        compiler = self.compiler.compiler_type
        if compiler in {'mingw32', 'unix'}:
            for e in self.extensions:
                e.define_macros = macros
                e.extra_compile_args = compile_args
                if compiler == 'mingw32':
                    e.extra_link_args = link_args_static
                else:
                    e.extra_link_args = link_args
        elif compiler == 'msvc':
            for e in self.extensions:
                e.define_macros = macros[:1]
                e.extra_compile_args = compile_args_msvc
                e.extra_link_args = link_args_msvc
        super(Build, self).build_extensions()


setup(ext_modules=ext_modules, cmdclass={'build_ext': Build})



# -*- coding: utf-8 -*-

from typing import Tuple, List, Iterable, Sequence, Callable, Optional, ClassVar
from enum import IntEnum, auto
from numpy import ndarray

# Color type must be a RGB data
_Color = Tuple[int, int, int]

def get_vlinks(vpoints: Iterable[VPoint]) -> List[VLink]:
    ...

class Coord:
    x: float
    y: float

    def __init__(self, x: float, y: float):
        """The constructor of Coordinate class."""
        ...

    def distance(self, p: Coord) -> float:
        ...

    def slope_angle(self, p: Coord) -> float:
        ...

    def is_nan(self) -> bool:
        ...

class VJoint(IntEnum):
    """Enumeration values of Joint types."""
    R = auto()
    P = auto()
    RP = auto()

class VPoint:
    links: Sequence[str]
    c: ndarray
    type: VJoint
    color: Optional[_Color]
    color_str: str
    type_str: str
    x: float
    y: float
    angle: float
    HOLDER: ClassVar[VPoint] = ...

    def __init__(
        self,
        links: Iterable[str],
        type_int: VJoint,
        angle: float,
        color_str: str,
        x: float,
        y: float,
        color_func: Optional[Callable[[str], _Color]] = None
    ):
        """The attributes will match to the object attributes of [VPoint] objects.

        Where the color function `color_func` needs to transform the color string `color_str` into RGB format.
        If color information is not needed, the `color_func` can be `None`.

        !!! note
            Some of the attributes are not work in some of the joint types.
        """
        ...

    @staticmethod
    def r_joint(links: Iterable[str], x: float, y: float) -> VPoint:
        ...

    @staticmethod
    def slider_joint(
        links: Iterable[str],
        type_int: VJoint,
        angle: float,
        x: float,
        y: float
    ) -> VPoint:
        ...

    def copy(self) -> VPoint:
        ...

    @property
    def sx(self) -> float:
        ...

    @property
    def sy(self) -> float:
        ...

    @property
    def cx(self) -> float:
        ...

    @property
    def cy(self) -> float:
        ...

    def set_links(self, links: Iterable[str]) -> None:
        ...

    def replace_link(self, link1: str, link2: str) -> None:
        ...

    def move(
        self,
        c1: Tuple[float, float],
        c2: Optional[Tuple[float, float]] = None
    ) -> None:
        ...

    def locate(self, x: float, y: float) -> None:
        ...

    def rotate(self, angle: float) -> None:
        ...

    def set_offset(self, offset: float) -> None:
        ...

    def disable_offset(self) -> None:
        ...

    def is_slider(self) -> bool:
        ...

    def distance(self, p: VPoint) -> float:
        ...

    def has_offset(self) -> bool:
        ...

    def offset(self) -> float:
        ...

    def true_offset(self) -> float:
        ...

    def slope_angle(self, p: VPoint, num1: int = 2, num2: int = 2) -> float:
        ...

    def link_pos(self, link: str) -> Coord:
        ...

    def grounded(self) -> bool:
        ...

    def pin_grounded(self) -> bool:
        ...

    def same_link(self, p: VPoint) -> bool:
        ...

    def no_link(self) -> bool:
        ...

    def is_slot_link(self, link: str) -> bool:
        ...

    def expr(self) -> str:
        ...

    def to_coord(self, ind: int) -> Coord:
        ...

    def __getitem__(self, i: int) -> float:
        ...

class VLink:
    name: str
    color_str: str
    color: Optional[_Color]
    points: Sequence[int]
    HOLDER: ClassVar[VLink] = ...
    FRAME: ClassVar[str] = ...

    def __init__(
        self,
        name: str,
        color_str: str,
        points: Iterable[int],
        color_func: Optional[Callable[[str], _Color]] = None
    ):
        """The attributes will match to the object attributes of [VLink] objects.

        Where the color function `color_func` needs to transform the color string `color_str` into RGB format.
        If color information is not needed, the `color_func` can be `None`.
        """
        ...

    def set_points(self, points: Iterable[int]) -> None:
        ...

    def points_pos(self, vpoints: Iterable[VPoint]) -> Sequence[Coord]:
        ...

    def __contains__(self, point: int) -> bool:
        ...



# -*- coding: utf-8 -*-
# cython: language_level=3, cdivision=True, boundscheck=False, wraparound=False
# cython: initializedcheck=False, nonecheck=False

"""Firefly Algorithm

author: Yuan Chang
copyright: Copyright (C) 2016-2021
license: AGPL
email: pyslvs@gmail.com
"""

cimport cython
from libc.math cimport exp
from .utility cimport uint, rand_v, ObjFunc, Algorithm


cdef double _distance(double[:] me, double[:] she, uint dim) nogil:
    """Distance of two fireflies."""
    cdef double dist = 0
    cdef uint i
    cdef double diff
    for i in range(dim):
        diff = me[i] - she[i]
        dist += diff * diff
    return dist


@cython.final
cdef class FA(Algorithm):
    """The implementation of Firefly Algorithm."""
    cdef double alpha, beta_min, beta0, gamma

    def __cinit__(
        self,
        ObjFunc func not None,
        dict settings not None,
        object progress_fun=None,
        object interrupt_fun=None
    ):
        # alpha, the step size
        self.alpha = settings['alpha']
        # beta_min, the minimal attraction, must not less than this
        self.beta_min = settings['beta_min']
        # beta0, the attraction of two firefly in 0 distance
        self.beta0 = settings['beta0']
        # gamma
        self.gamma = settings['gamma']

    cdef inline void move_fireflies(self) nogil:
        """Move fireflies."""
        cdef bint is_move
        cdef uint i, j, s
        for i in range(self.pop_num):
            moved = False
            for j in range(self.pop_num):
                if i == j or self.fitness[i] <= self.fitness[j]:
                    continue
                self.move_firefly(self.pool[i, :], self.pool[j, :])
                moved = True
            if moved:
                continue
            # Evaluate
            for s in range(self.dim):
                self.pool[i, s] = self.check(s, self.pool[i, s] + self.alpha * (
                    self.func.ub[s] - self.func.lb[s]) * rand_v(-0.5, 0.5))

    cdef inline void move_firefly(self, double[:] me, double[:] she) nogil:
        """Move single firefly."""
        cdef double r = _distance(me, she, self.dim)
        cdef double beta = ((self.beta0 - self.beta_min)
                            * exp(-self.gamma * r) + self.beta_min)
        cdef uint s
        for s in range(self.dim):
            me[s] = self.check(s, me[s] + beta * (she[s] - me[s]) + self.alpha
                               * (self.func.ub[s] - self.func.lb[s])
                               * rand_v(-0.5, 0.5))

    cdef inline void generation(self) nogil:
        self.move_fireflies()
        # Get fitness
        cdef uint i
        for i in range(self.pop_num):
            self.fitness[i] = self.func.fitness(self.pool[i, :])
        self.find_best()
